@@ -26,4 +26,4 @@ from lined.tools import (
     side_call,
     print_and_pass_on,
 )
-from lined.util import partial_plus
+from lined.util import partial_plus, n_required_args, dot_to_ascii
